/*
 * Copyright 2016 - 2024 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <ox/std/array.hpp>
#include <ox/std/point.hpp>
#include <ox/std/size.hpp>
#include <ox/std/types.hpp>
#include <ox/std/vector.hpp>
#include <ox/model/def.hpp>

#include <turbine/turbine.hpp>

#include "color.hpp"

namespace nostalgia::core {

struct NostalgiaPalette {
	static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.NostalgiaPalette";
	static constexpr auto TypeVersion = 1;
	ox::Vector<Color16> colors = {};
};

struct PaletteV1 {
	static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.Palette";
	static constexpr auto TypeVersion = 1;
	ox::Vector<Color16> colors;
};

struct PaletteV2 {
	static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.Palette";
	static constexpr auto TypeVersion = 2;
	static constexpr auto Preloadable = true;
	ox::Vector<ox::Vector<Color16>> pages;
};

struct PaletteV3 {
	static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.Palette";
	static constexpr auto TypeVersion = 3;
	static constexpr auto Preloadable = true;
	struct ColorInfo {
		static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.Palette.ColorInfo";
		static constexpr auto TypeVersion = 3;
		ox::String name;
		constexpr ColorInfo() noexcept = default;
		constexpr explicit ColorInfo(ox::StringView pName) noexcept:
			name(pName) {}
		constexpr explicit ColorInfo(ox::String &&pName) noexcept:
			name(std::move(pName)) {}
	};
	ox::Vector<ColorInfo> colorInfo;
	ox::Vector<ox::Vector<Color16>> pages;
};

using Palette = PaletteV3;


struct CompactPaletteV1 {
	static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.CompactPalette";
	static constexpr auto TypeVersion = 1;
	static constexpr auto Preloadable = true;
	ox::Vector<ox::Vector<Color16>> pages{};
};

using CompactPalette = CompactPaletteV1;

[[nodiscard]]
constexpr bool valid(Palette const&p) noexcept {
	auto const colors = p.colorInfo.size();
	return ox::all_of(p.pages.begin(), p.pages.end(), [colors](auto const&page) {
		return page.size() == colors;
	});
}

[[nodiscard]]
constexpr Color16 color(Palette const&pal, size_t page, size_t idx) noexcept {
	if (page < pal.pages.size() && idx < pal.pages[page].size()) [[likely]] {
		return pal.pages[page][idx];
	}
	return 0;
}

[[nodiscard]]
constexpr Color16 color(CompactPalette const&pal, size_t page, size_t idx) noexcept {
	if (page < pal.pages.size() && idx < pal.pages[page].size()) [[likely]] {
		return pal.pages[page][idx];
	}
	return 0;
}

[[nodiscard]]
constexpr Color16 color(Palette const&pal, size_t idx) noexcept {
	return color(pal, 0, idx);
}

[[nodiscard]]
constexpr Color16 color(CompactPalette const&pal, size_t idx) noexcept {
	return color(pal, 0, idx);
}

[[nodiscard]]
constexpr auto &colors(Palette &pal, size_t page = 0) noexcept {
	return pal.pages[page];
}

[[nodiscard]]
constexpr auto &colors(CompactPalette &pal, size_t page = 0) noexcept {
	return pal.pages[page];
}

[[nodiscard]]
constexpr auto &colors(Palette const&pal, size_t page = 0) noexcept {
	return pal.pages[page];
}

[[nodiscard]]
constexpr auto &colors(CompactPalette const&pal, size_t page = 0) noexcept {
	return pal.pages[page];
}

[[nodiscard]]
constexpr size_t colorCnt(Palette const&pal, size_t page = 0) noexcept {
	if (page < pal.pages.size()) [[likely]] {
		return pal.pages[page].size();
	}
	return 0;
}

[[nodiscard]]
constexpr size_t colorCnt(CompactPalette const&pal, size_t page = 0) noexcept {
	if (page < pal.pages.size()) [[likely]] {
		return pal.pages[page].size();
	}
	return 0;
}

[[nodiscard]]
constexpr size_t largestPage(Palette const&pal) noexcept {
	size_t out{};
	for (auto const&page : pal.pages) {
		out = ox::max(out, page.size());
	}
	return out;
}

[[nodiscard]]
constexpr size_t largestPage(CompactPalette const&pal) noexcept {
	size_t out{};
	for (auto const&page : pal.pages) {
		out = ox::max(out, page.size());
	}
	return out;
}

oxModelBegin(NostalgiaPalette)
	oxModelField(colors)
oxModelEnd()

oxModelBegin(PaletteV1)
	oxModelField(colors)
oxModelEnd()

oxModelBegin(PaletteV2)
	oxModelField(pages)
oxModelEnd()

oxModelBegin(PaletteV3::ColorInfo)
	oxModelField(name)
oxModelEnd()

oxModelBegin(PaletteV3)
	oxModelField(colorInfo)
	oxModelField(pages)
oxModelEnd()

oxModelBegin(CompactPaletteV1)
	oxModelField(pages)
oxModelEnd()

}
